%This function handles the training or testing of a NIMBLE model
function [Model, Stats, Cache] = Train_And_Test_Model(Model, files, operating_mode, max_samples, class_names, img_size, verbose, Cache)

model_eval_timer = tic();
if isfield(Model, 'parallel_mode') && Model.parallel_mode
    Start_MATLAB_Pool();
end

%set up feature caching
if ~exist('verbose', 'var')
    verbose = true;
end

Stats = [];

num_classes = size(files, 1);
Correct_Per_Class = zeros(1, num_classes);
Total_Per_Class = zeros(1, num_classes);

if operating_mode == 1 %training
    Model.PCA_Vectors = [];
    Model.PCA_Mean = [];
    Model.Net = [];
elseif operating_mode == 2 %test
    Confusion_Matrix = zeros(num_classes, num_classes, 'single');
end


if operating_mode == 1
    num_samples = max_samples;
else
    num_samples = max_samples;
end

%this is a dummy call used to allocate memory
[Patch_Collection, Location_Collection, ~, ~, ~, Model] = Process_Samples(operating_mode, rand(Model.img_size, Model.img_size, 3, 'single'), 0, num_samples, Model);

Patch_Collection_Size = size(Patch_Collection, 2);
max_data = size(Patch_Collection, 2)*numel(files);
if operating_mode == 1
    training_data = zeros(size(Patch_Collection, 1), max_data, 'single');
    training_locations = zeros(size(Location_Collection, 1), max_data, 'single');
    training_labels = zeros(1, max_data, 'single');
end


%class_list = randperm(size(files, 1));
class_list = 1:size(files, 1);

Mean_AUC = 0;
data_start = 1;
Correct_Sum = zeros(num_classes, max_samples, 'single');
Class_Prob_History = cell(size(files, 1), 1);
%loop for each category
for class = 1:size(files, 1)
    num_instances = 0;
    
    C_Class_Prob = zeros(size(files, 2), size(files, 1));
    
    for lo = 1:size(files, 2)
        if ~isempty(files{class, lo})
            num_instances = num_instances + 1;
        else
            break;
        end
    end
    
    Correct_Sum_Result = 0;
    
    instance_data = cell(1, num_instances);
    valid_instances = false(1, num_instances);
    instance_labels = cell(1, num_instances);
    instance_locations = cell(1, num_instances);
        
    
    Correct_Per_Class_i = 0;
    Total_Per_Class_i = 0;
    
    Confusion_Vector = zeros(num_instances, 1);
    parfor instance = 1:num_instances
    %for instance = 1:num_instances %can make this a parfor
        img = [];
        if ~isempty(files{class, instance})
            try
                img = imread(files{class, instance});               
            catch
                img = [];
                fprintf('Unable to process the file: %s\n', files{class, instance});
            end
        end
        
        
        if ~isempty(img)
            valid_instances(instance) = true;
            
            %get samples
            [Patch_Collection, Location_Collection, a, Correct_Time, Class_Prob] = Process_Samples(operating_mode, img, class, num_samples, Model);
                                              
            Correct_Time(~isfinite(Correct_Time)) = 0;
            Correct_Sum_Result = Correct_Sum_Result + Correct_Time;
            
            
            %Store data if training
            if operating_mode == 1
                %store if training
                instance_data{instance} = Patch_Collection;
                instance_labels{instance} = class*ones(1, Patch_Collection_Size, 'single');               
                instance_locations{instance} = Location_Collection;
            end
            
            if operating_mode == 2
                %evaluate if testing
                if length(Class_Prob) > 1
                    C_Class_Prob(instance, :) = Class_Prob;
                    [c_val, c_ndx] = max(Class_Prob);
                    
                    Correct_Per_Class_i = Correct_Per_Class_i + (class == c_ndx);
                    Total_Per_Class_i = Total_Per_Class_i + 1;
                    Confusion_Vector(instance) = c_ndx;
                end
            end
        end
    end
        
    if operating_mode == 1
        %store away training data
        instance_data(~valid_instances) = [];
        temp_data = cell2mat(instance_data);        
        data_end = (data_start - 1) + size(temp_data, 2);
        training_data(:, data_start:data_end) = temp_data;
        temp_data = cell2mat(instance_locations);
        training_locations(:, data_start:data_end) = temp_data;
        training_labels(data_start:data_end) = class;       
        data_start = data_end + 1;        
    end
    
    try
        Correct_Sum(class, :) = Correct_Sum_Result;
    catch
    end
    
    if operating_mode == 2
        %generate confusion matrices after testing
        Class_Prob_History{class} = C_Class_Prob;
        
        for instance = 1:num_instances
            try
                Confusion_Matrix(Confusion_Vector(instance), class) = Confusion_Matrix(Confusion_Vector(instance), class) + 1;
            end
        end
        
        Correct_Per_Class(class) = Correct_Per_Class_i;
        Total_Per_Class(class) =  Total_Per_Class_i;
        
        if Total_Per_Class(class) == 0
            AUC = 0;
        else
            curv = Correct_Sum(class, :) ./ Total_Per_Class(class);
            
            %Area under performance as a function of fixations curve
            AUC = (trapz(curv) / (max_samples)) * 100;
        end
        Mean_AUC = Mean_AUC + AUC;
        if verbose
            %AUC has little to do with ROC
            if isempty(class_names)
                fprintf('Class %d Accuracy: %.2f  (AUC: %.2f)\n', class, 100*(Correct_Per_Class(class) ./ Total_Per_Class(class)), AUC);
            else
                fprintf('Class %d (%s) Accuracy: %.2f  (AUC: %.2f)\n', class, class_names{class}, 100*(Correct_Per_Class(class) ./ Total_Per_Class(class)), AUC);
            end
        end
    end
end

class = class_list(end);
if operating_mode == 1
    clear instance_data;
    clear files;
    clear Patch_Collection;
    
    if data_end < size(training_data, 2)
        training_data = training_data(:, 1:data_end);
    end
    
    Stats.training_data = training_data;
    Stats.training_labels = training_labels(1:data_end);
    Stats.training_locations = training_locations(:, 1:data_end);
end

if operating_mode == 2
    %Compute remaining test statistics
    Total_Per_Class(Total_Per_Class == 0) = 1;
    accuracy = mean(Correct_Per_Class(1:class) ./ Total_Per_Class(1:class));
    
    Stats.Temporal_Data = bsxfun(@rdivide, Correct_Sum(1:class, :), Total_Per_Class(1:class)');
    Temporal_Accuracy = 100*mean(Stats.Temporal_Data, 1);
    
    Stats.Final_Accuracy = 100*sum(Correct_Per_Class) ./ sum(Total_Per_Class);
    Stats.Mean_Class_Accuracy = mean(accuracy) * 100;
    
    Stats.Temporal_Accuracy = Temporal_Accuracy;
    Stats.Max_Accuracy = max(Temporal_Accuracy);
    
    Stats.Class_Prob_History = Class_Prob_History;
    Stats.Mean_AUC = Mean_AUC / class;
    for col = 1:size(Confusion_Matrix, 1);
        Confusion_Matrix(:, col) = Confusion_Matrix(:, col) / sum(Confusion_Matrix(:, col));
    end
    Stats.Confusion_Matrix = Confusion_Matrix;
    if verbose
        fprintf('Final Mean Per Class Accuracy: %.2f\n', Stats.Mean_Class_Accuracy);
        fprintf('Final Accuracy: %.2f\n', Stats.Final_Accuracy);
        fprintf('Max Per Class Accuracy: %.2f\n', max(Temporal_Accuracy));
        %fprintf('Mean AUC: %.2f\n', Stats.Mean_AUC);
    end
end


took = toc(model_eval_timer);
fprintf('Took: %.2f minutes\n', took / 60);